#include "SSP.h"

DispatchBatchSSP::DispatchBatchSSP(FABMAS_SB_ToolGroup* tg, string name) : FABMAS_SB_SSP(tg, name)
{
}

void DispatchBatchSSP::initialize_problem() {
  FABMAS_SB_SSP::initialize_problem();

  // calculate batchable tasks
  list<FABMAS_SB_Task*>::const_iterator i = schedulable_tasks.begin();
  while (i != schedulable_tasks.end()) {
    FABMAS_SB_Task* task = *i;
    sorted_tasks_by_family[task->get_setup()].push_back(task);
    i++;
  }
}

void DispatchBatchSSP::solve_problem() {
  while (!sorted_tasks_by_family.empty()) {
    FABMAS_SB_Tool* tool = toolgroup->calculate_next_avail_tool();
    FABMAS_SB_Date tool_avail_time = tool->get_avail_time();
    FABMAS_SB_Date min_avail_time = 0;
    FABMAS_SB_Task* choosed_task = 0; 
    FABMAS_SB_Task* choosed_task_late = 0;

    // calculate_dynamic_priorities(tool, schedulable_tasks);
    map<FABMAS_SB_Setup*, list<FABMAS_SB_Task*> >::iterator i = sorted_tasks_by_family.begin();
    while (i != sorted_tasks_by_family.end()) {
      FABMAS_SB_Setup* bg    = (*i).first;
      list<FABMAS_SB_Task*> tasks = (*i).second;

      // get the first task
      FABMAS_SB_Task* task = get_earliest_available_task(tool, tasks);
      if (task != 0) {
	FABMAS_SB_Date task_avail_time = task->calculate_avail_time();

	if (task_avail_time <= tool_avail_time) {
	  if (better_task(task, choosed_task)) {
	    choosed_task   = task;
	    min_avail_time = tool_avail_time;
	  }
	} else {
	  if (choosed_task_late == 0) {
	    // These are the nodes that are available at a later time than the current time
	    // If we havn't yet found any node with an availabiltiy time <= currentTime, determine
	    if ((task_avail_time < min_avail_time) || (choosed_task_late == 0) ||
		((task_avail_time == min_avail_time) && better_task(task, choosed_task_late)))
	      {
		min_avail_time     = task_avail_time;
		choosed_task_late  = task;
	      }
	  }
	}
      }

      i++;
    }

    if (choosed_task == 0) choosed_task = choosed_task_late;

    if (choosed_task != 0) {
      FABMAS_SB_Date batch_start_time = max(tool_avail_time, min_avail_time);
      list<FABMAS_SB_Task*>& tasks = sorted_tasks_by_family[choosed_task->get_setup()];

      FABMAS_SB_Batch batch = build_batch(tasks, batch_start_time, tool);
      filter_tasks(sorted_tasks_by_family[choosed_task->get_setup()], batch.tasks);
      if (sorted_tasks_by_family[choosed_task->get_setup()].empty()) 
	sorted_tasks_by_family.erase(choosed_task->get_setup());
			
      filter_tasks(schedulable_tasks, batch.tasks);

      tool->add_to_schedule(batch.tasks);
    } else {
      list<FABMAS_SB_Task*>::iterator i = schedulable_tasks.begin();
      while (i != schedulable_tasks.end()) {
	FABMAS_SB_Task* task = *i;
	task->mark_as_unschedulable();
	i++;
      }

      sorted_tasks_by_family.clear();
      schedulable_tasks.clear();
    }
  }
}

FABMAS_SB_Batch DispatchBatchSSP::build_batch(const list<FABMAS_SB_Task*>& tasks, FABMAS_SB_Date current_time, FABMAS_SB_Tool* tool)
{
  FABMAS_SB_Batch batch; bool finished = false;
  while ((!tasks.empty()) && (!finished)) {
    FABMAS_SB_Task* task = calculate_next_available_task(tasks);
		
    if (task != 0) {
      if (batch.tasks.size() < tool->get_batch_min()) {
	batch.tasks.insert(task); task->set_scheduled_tool(tool);
	current_time = max(current_time, task->calculate_avail_time());
      } else {
	if (batch.tasks.size() < tool->get_batch_max()) {
	  if (task->calculate_avail_time() <= current_time) {
	    batch.tasks.insert(task); task->set_scheduled_tool(tool);
	  } else {
	    finished = true;
	  }
	} else {
	  finished = true;
	}
      }
    } else {
      finished = true;
    }
  }
	
  batch.set_scheduled_tool(0);
  return batch;
}
